/* vi: set sw=4 ts=4: */
/*
 * prioritynames[] and facilitynames[]
 *
 * Copyright (C) 2008 by Denys Vlasenko <vda.linux@gmail.com>
 *
 * Licensed under GPLv2, see file LICENSE in this source tree.
 */

#include "libbb.h"
#define SYSLOG_NAMES
#define SYSLOG_NAMES_CONST
#include <syslog.h>

#if 1
/* For the record: with SYSLOG_NAMES <syslog.h> defines
 * (not declares) the following:
 */
typedef struct _code {
	/*const*/ char *c_name;
	int c_val;
} CODE;
/*const*/ CODE prioritynames[] = {
    { "alert", LOG_ALERT },
    { "crit",       LOG_CRIT },
    { "debug",      LOG_DEBUG },
    { "emerg",      LOG_EMERG },
    { "err",        LOG_ERR },
    { "error",      LOG_ERR },              /* DEPRECATED */
    { "info",       LOG_INFO },
//    { "none",       INTERNAL_NOPRI },       /* INTERNAL */
    { "notice",     LOG_NOTICE },
    { "panic",      LOG_EMERG },            /* DEPRECATED */
    { NULL, -1 }
};
/*const*/ CODE facilitynames[] = {
    { "auth",       LOG_AUTH },
    { "authpriv",   LOG_AUTHPRIV },
    { "cron",       LOG_CRON },
    { "daemon",     LOG_DAEMON },
    { "ftp",        LOG_FTP },
    { "kern",       LOG_KERN },
    { "lpr",        LOG_LPR },
    { "mail",       LOG_MAIL },
  //  { "mark",       INTERNAL_MARK },        /* INTERNAL */
    { "news",       LOG_NEWS },
};

/* This MUST occur only once per entire executable,
 * therefore we can't just do it in syslogd.c and logger.c -
 * there will be two copies of it.
 *
 * We cannot even do it in separate file and then just reference
 * prioritynames[] from syslogd.c and logger.c - bare <syslog.h>
 * will not emit extern decls for prioritynames[]! Attempts to
 * emit "matching" struct _code declaration defeat the whole purpose
 * of <syslog.h>.
 *
 * For now, syslogd.c and logger.c are simply compiled into
 * one object file.
 */
#endif

#if ENABLE_SYSLOGD
#include "syslogd.c"
#endif

#if ENABLE_LOGGER
#include "logger.c"
#endif
